function initialdata!(env, P)
  id = P.prmsdb["GRHD"]["id"]
  refsol = P.prmsdb["GRHD"]["analyze_error_reference_solution"]
  if !isempty(refsol)
    if refsol == "from_id"
      refsol = string(id)
    end
    initialdata_equation!(Val(Symbol(refsol)), env, P, P.prmsdb["GRHD"])
    for vars in (P.prmsdb["GRHD"]["variables0d_analyze_error"],
                 P.prmsdb["GRHD"]["variables1d_analyze_error"])
      for var in vars
        var_ref = string(var, "_ref")
        v = get_variable(env.mesh.cache, Symbol(var))
        v_ref = get_variable(env.mesh.cache, Symbol(var_ref))
        @. v_ref = v
      end
    end
  end
  initialdata_equation!(Val(Symbol(id)), env, P, P.prmsdb["GRHD"])
  # # initialdata_entropy_variables!(id, P, prms)
  initialdata_hrsc!(Val(Symbol(id)), env, P, P.hrsc, P.prmsdb["GRHD"])
  # initialdata_tci!(Val(Symbol(id)), env, P, P.tci, prms)
  if P.prmsdb["GRHD"]["training_wheels"]
    initialdata_smooth!(env, P, P.prmsdb["GRHD"])
  end
end


#######################################################################
#                     Initial data gravity sector                     #
#######################################################################


#######################################################################
#                        Bondi accretion tests                        #
#######################################################################


include("spherical_accretion.jl")


# 'Online' computation of Bondi data with BondiAccretion.solve is preferred
# over using precomputed data and reading them with this function.
# Reason is that the latter suffers from interpolation errors,
# in fact, we need the online method to be able to do proper convergence tests.
function readdata_bondi(id_filename)
  if !isfile(id_filename)
    error("GRHD: Failed to locate bondi_accretion initial data at $id_filename")
  end

  data = h5open(id_filename, "r") do file
    g = file["bondi_accretion"]
    Γ  = read_attribute(g, "Γ")
    K  = read_attribute(g, "K")
    ρc = read_attribute(g, "ρc")
    rc = read_attribute(g, "rc")
    M  = read_attribute(g, "M")
    r  = read(g, "r")
    ρ  = read(g, "ρ")
    p  = read(g, "p")
    ϵ  = read(g, "ϵ")
    vr = read(g, "vr")
    (; Γ, K, ρc, rc, M, r, ρ, ϵ, p, vr)
  end

  return data
end


function initialdata_equation!(::Val{:bondi_accretion}, env, P::Project{:valencia1d}, prms)

  @unpack mesh = env
  @unpack equation = P

  @assert data.K ≈ equation.eos.K
  @assert data.Γ ≈ equation.eos.Gamma

  @unpack rD, rSr, rτ = get_dynamic_variables(mesh.cache)
  @unpack D, Sr, τ, ρ, p, ϵ, vr, r, max_v,
          init_D, init_Sr, init_τ, init_vr, init_p, init_ρ, init_ϵ,
          init_max_v = get_static_variables(mesh.cache)
  @unpack α, βru, ∂αr, ∂βrrdu, sqrt_detγ,
          γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
          Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕrϕ, Γϕθϕ,
          Krr, Kθθ, Kϕϕ = get_static_variables(mesh.cache)

  # reference data from bugner's solver SphericalAccretion.c (extracted from bamps)
  M = 1.0
  data = SphericalAccretion.solve(; rs=r, rc=200*M, ρc=7e-4, Γ=4/3, M=M)
  @assert data.K ≈ equation.eos.K
  @assert data.Γ ≈ equation.eos.Gamma
  p   = data.ps
  ρ   = data.ρs
  ϵ   = data.ϵs
  vru = data.vrs

  equation.atm_density = maximum(ρ) * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  # to avoid problems with 1/sin(θ) etc., this is also done in initialdata/SphericalAccretion.jl
  θ = π/2
  @unpack M = data

  @. α   = sqrt(r/(r+2*M))
  @. βru = 2*M/(r+2*M) # = βʳ
  @. ∂αr = M / sqrt(r*(2*M+r)^3) # = ∂ᵣα
  @. ∂βrrdu = - 2*M/(2*M+r)^2 # = ∂ᵣβʳ

  # γ_ij
  @. γrr = 1 + 2*M/r
  @. γθθ = r^2
  @. γϕϕ = r^2 * sin(θ)^2
  @. sqrt_detγ = sqrt(γrr*γθθ*γϕϕ)

  # # γijuu = γ^ij
  # γrruu = @. 1/γrr
  # γθθuu = @. 1/γθθ
  # γϕϕuu = @. 1/γϕϕ

  # ∂iγjk := ∂γijk
  @. ∂γrrr = -2*M/r^2
  @. ∂γrθθ = 2*r
  @. ∂γrϕϕ = 2*r*sin(θ)^2
  @. ∂γθϕϕ = r^2 * sin(θ)*cos(θ)

  # Γijk := Γ^i_jk wrt γ_ij
  @. Γrrr = -M/(r*(2*M+r))
  @. Γrθθ = -r^2/(2*M+r)
  @. Γrϕϕ = -r^2*sin(θ)^2/(2*M+r)
  @. Γθrθ = 1/r
  @. Γθϕϕ = -sin(θ)*cos(θ)
  @. Γϕrϕ = 1/r
  @. Γϕθϕ = cos(θ)/sin(θ)

  # Kij := K_ij
  @. Krr = -2*M*(M+r)/sqrt(r^5*(2*M+r))
  @. Kθθ = 2*M*sqrt(r/(2*M+r))
  @. Kϕϕ = 2*M*sqrt(r/(2*M+r))*sin(θ)^2

  # # ∂Kijk = ∂_i K_jk
  # ∂Krrr = @. 2*M*(5*M^2+2*r^2+6*M*r)/sqrt(r^7*(2*M+r)^3)
  # ∂Krθθ = @. 2*M^2/sqrt(r*(2*M+r)^3)
  # ∂Krϕϕ = @. ∂Krθθ * sin(θ)^2

  # # verify trace of K
  # trK_v1 = @. Krr*γrruu + Kθθ*γθθuu + Kϕϕ*γϕϕuu
  # trK_v2 = @. 2*M*α^3/r^2*(1+3*M/r)
  # @show(extrema(trK_v1.-trK_v2))


  # from initialdata/SphericalAccretion.jl
  # also see Papadopoulos and Font 1998, arXiv:gr-qc/9803087, eq (5)
  v2  = @. γrr*vru^2
  W   = @. 1/sqrt(1-v2)  # = αu^t, Lorentz factor
  @. vr = γrr * vru # = γ_rr v^r

  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D

  if P.prmsdb["GRHD"]["c2p_enforce_causal_atm"]
    broadcast_volume!(determine_atmosphere, equation, mesh)
  else
    @unpack c2p_freeze_atm = get_static_variables(mesh)
    c2p_freeze_atm .= 0.0 # atmosphere can always change
  end
  broadcast_volume!(cons2prim_valencia1d, equation, mesh)
  broadcast_volume!(maxspeed_valencia1d, equation, mesh)

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v
  @. init_ρ     = ρ
  @. init_ϵ     = ϵ

  return
end


function initialdata_equation!(::Val{:bondi_accretion}, env, P::Project{:spherical1d}, prms)

  @unpack mesh = env
  @unpack equation = P

  @unpack D, Sr, τ = get_dynamic_variables(mesh.cache)
  @unpack r, max_v,
          init_D, init_Sr, init_τ, init_vr, init_p, init_ρ, init_ϵ,
          init_max_v = get_static_variables(mesh.cache)
  @unpack α, βru, ∂αr, ∂βrrdu,
          γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
          Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕrϕ, Γϕθϕ,
          Krr, Kθθ, Kϕϕ = get_static_variables(mesh.cache)

  # reference data from bugner's solver SphericalAccretion.c (extracted from bamps)
  M = 1.0
  data = SphericalAccretion.solve(; rs=r, rc=200*M, ρc=7e-4, Γ=4/3, M=M)
  @assert data.K ≈ equation.eos.K
  @assert data.Γ ≈ equation.eos.Gamma
  # not storing in cached primitives, because they will be computed consistently with the c2p below
  p   = data.ps
  ρ   = data.ρs
  ϵ   = data.ϵs
  vru = data.vrs

  equation.atm_density = maximum(ρ) * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  # to avoid problems with 1/sin(θ) etc., this is also done in initialdata/SphericalAccretion.jl
  θ = π/2
  @unpack M = data

  @. α   = sqrt(r/(r+2*M))
  @. βru = 2*M/(r+2*M) # = β^r
  @. ∂αr = M / sqrt(r*(2*M+r)^3)
  @. ∂βrrdu = - 2*M/(2*M+r)^2

  # γ_ij
  @. γrr = 1 + 2*M/r
  @. γθθ = r^2
  @. γϕϕ = r^2 * sin(θ)^2

  # # γijuu = γ^ij
  # γrruu = @. 1/γrr
  # γθθuu = @. 1/γθθ
  # γϕϕuu = @. 1/γϕϕ

  # ∂iγjk := ∂γijk
  @. ∂γrrr = -2*M/r^2
  @. ∂γrθθ = 2*r
  @. ∂γrϕϕ = 2*r*sin(θ)^2
  @. ∂γθϕϕ = r^2 * sin(θ)*cos(θ)

  # Γijk := Γ^i_jk wrt γ_ij
  @. Γrrr = -M/(r*(2*M+r))
  @. Γrθθ = -r^2/(2*M+r)
  @. Γrϕϕ = -r^2*sin(θ)^2/(2*M+r)
  @. Γθrθ = 1/r
  @. Γθϕϕ = -sin(θ)*cos(θ)
  @. Γϕrϕ = 1/r
  @. Γϕθϕ = cos(θ)/sin(θ)

  # Kij := K_ij
  @. Krr = -2*M*(M+r)/sqrt(r^5*(2*M+r))
  @. Kθθ = 2*M*sqrt(r/(2*M+r))
  @. Kϕϕ = 2*M*sqrt(r/(2*M+r))*sin(θ)^2

  # # ∂Kijk = ∂_i K_jk
  # ∂Krrr = @. 2*M*(5*M^2+2*r^2+6*M*r)/sqrt(r^7*(2*M+r)^3)
  # ∂Krθθ = @. 2*M^2/sqrt(r*(2*M+r)^3)
  # ∂Krϕϕ = @. ∂Krθθ * sin(θ)^2

  # # verify trace of K
  # trK_v1 = @. Krr*γrruu + Kθθ*γθθuu + Kϕϕ*γϕϕuu
  # trK_v2 = @. 2*M*α^3/r^2*(1+3*M/r)
  # @show(extrema(trK_v1.-trK_v2))


  # from initialdata/SphericalAccretion.jl
  # also see Papadopoulos and Font 1998, arXiv:gr-qc/9803087, eq (5)
  v2  = @. γrr*vru^2
  W   = @. 1/sqrt(1-v2)  # = αu^t, Lorentz factor
  vr  = @. γrr * vru # = γ_rr v^r

  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D

  if P.prmsdb["GRHD"]["c2p_enforce_causal_atm"]
    broadcast_volume!(determine_atmosphere, equation, mesh)
  else
    @unpack c2p_freeze_atm = get_static_variables(mesh)
    c2p_freeze_atm .= 0.0 # atmosphere can always change
  end
  broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  # consistency check
  let ref_ρ=ρ
    @unpack ρ = get_static_variables(mesh)
    err = maximum(abs.(ref_ρ.-ρ))
    max_ρ = maximum(abs, ref_ρ)
    if err/max_ρ > 1e-6
      error("setup of bondi_accretion data failed")
    end
  end

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v
  @. init_ρ     = ρ
  @. init_ϵ     = ϵ

  return
end


function initialdata_equation!(::Val{:bondi_accretion_infall}, env, P::Project{:spherical1d}, prms)

  @unpack mesh = env
  @unpack equation = P

  @unpack D, Sr, τ = get_dynamic_variables(mesh.cache)
  @unpack r, max_v,
          init_D, init_Sr, init_τ, init_vr, init_p, init_ρ, init_ϵ,
          init_max_v = get_static_variables(mesh.cache)
  @unpack α, βru, ∂αr, ∂βrrdu,
          γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
          Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕrϕ, Γϕθϕ,
          Krr, Kθθ, Kϕϕ = get_static_variables(mesh.cache)

  # reference data from Bugner's solver SphericalAccretion.c (extracted from bamps)
  M = 1.0
  data = SphericalAccretion.solve(; rs=r, rc=200*M, ρc=7e-4, Γ=4/3, M=M)
  @assert data.K ≈ equation.eos.K
  @assert data.Γ ≈ equation.eos.Gamma
  # below this radius we impose a low density fluid using the same EoS as for the static solution
  rs = P.prmsdb["GRHD"]["id_bondi_accretion_infall_rs"]
  rmin, rmax = extrema(r)
  @assert rmin < rs < rmax
  idx_interior = findfirst(>(rs), r)
  interior = 1:idx_interior
  # not storing in cached primitives, because they will be computed consistently with the c2p below
  ρ   = data.ρs
  p   = data.ps
  ϵ   = data.ϵs
  vru = copy(data.vrs)

  # density chosen according to Bugner's thesis
  ρl = P.prmsdb["GRHD"]["id_bondi_accretion_infall_ρl"]
  @. ρ[interior] = ρl
  @. p[interior] = equation.eos(Pressure, Density, ρ[interior])
  @. ϵ[interior] = equation.eos(Pressure, Density, ρ[interior])
  @. vru[interior] = 0.0

  equation.atm_density = maximum(ρ) * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  # to avoid problems with 1/sin(θ) etc., this is also done in initialdata/SphericalAccretion.jl
  θ = π/2
  @unpack M = data

  @. α   = sqrt(r/(r+2*M))
  @. βru = 2*M/(r+2*M) # = β^r
  @. ∂αr = M / sqrt(r*(2*M+r)^3)
  @. ∂βrrdu = - 2*M/(2*M+r)^2

  # γ_ij
  @. γrr = 1 + 2*M/r
  @. γθθ = r^2
  @. γϕϕ = r^2 * sin(θ)^2

  # # γijuu = γ^ij
  # γrruu = @. 1/γrr
  # γθθuu = @. 1/γθθ
  # γϕϕuu = @. 1/γϕϕ

  # ∂iγjk := ∂γijk
  @. ∂γrrr = -2*M/r^2
  @. ∂γrθθ = 2*r
  @. ∂γrϕϕ = 2*r*sin(θ)^2
  @. ∂γθϕϕ = r^2 * sin(θ)*cos(θ)

  # Γijk := Γ^i_jk wrt γ_ij
  @. Γrrr = -M/(r*(2*M+r))
  @. Γrθθ = -r^2/(2*M+r)
  @. Γrϕϕ = -r^2*sin(θ)^2/(2*M+r)
  @. Γθrθ = 1/r
  @. Γθϕϕ = -sin(θ)*cos(θ)
  @. Γϕrϕ = 1/r
  @. Γϕθϕ = cos(θ)/sin(θ)

  # Kij := K_ij
  @. Krr = -2*M*(M+r)/sqrt(r^5*(2*M+r))
  @. Kθθ = 2*M*sqrt(r/(2*M+r))
  @. Kϕϕ = 2*M*sqrt(r/(2*M+r))*sin(θ)^2

  # # ∂Kijk = ∂_i K_jk
  # ∂Krrr = @. 2*M*(5*M^2+2*r^2+6*M*r)/sqrt(r^7*(2*M+r)^3)
  # ∂Krθθ = @. 2*M^2/sqrt(r*(2*M+r)^3)
  # ∂Krϕϕ = @. ∂Krθθ * sin(θ)^2

  # # verify trace of K
  # trK_v1 = @. Krr*γrruu + Kθθ*γθθuu + Kϕϕ*γϕϕuu
  # trK_v2 = @. 2*M*α^3/r^2*(1+3*M/r)
  # @show(extrema(trK_v1.-trK_v2))


  # from initialdata/SphericalAccretion.jl
  # also see Papadopoulos and Font 1998, arXiv:gr-qc/9803087, eq (5)
  v2  = @. γrr*vru^2
  W   = @. 1/sqrt(1-v2)  # = αu^t, Lorentz factor
  vr = @. γrr * vru # = γ_rr v^r

  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D

  if P.prmsdb["GRHD"]["c2p_enforce_causal_atm"]
    broadcast_volume!(determine_atmosphere, equation, mesh)
  else
    @unpack c2p_freeze_atm = get_static_variables(mesh)
    c2p_freeze_atm .= 0.0 # atmosphere can always change
  end
  broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v
  @. init_ρ     = ρ
  @. init_ϵ     = ϵ

  let ref_ρ=ρ
    @unpack ρ = get_static_variables(mesh)
    if abs(ref_ρ[end]-ρ[end])/ρ[end] > 1e-6
      error("setup of bondi_accretion_infall data failed")
    end
  end

  return
end


#######################################################################
#                                 TOV                                 #
#######################################################################


function readdata_tov(filename)
  if !isfile(filename)
    error("GRHD: Failed to locate TOV initial data at id_filename = $filename")
  end

  data = h5open(filename, "r") do file
    g = file["TOV"]
    Γ  = read_attribute(g, "Γ")
    K  = read_attribute(g, "K")
    ρc = read_attribute(g, "ρc")
    M  = read_attribute(g, "M")
    rsch = read(g, "r") # schwarzschild radius coordinate
    riso = read(g, "R") # isotropic radius coordinate
    ρ   = read(g, "ρ")
    p   = read(g, "p")
    ϵ   = read(g, "ϵ")
    ϕ   = read(g, "ϕ")
    m   = read(g, "m")

    # metric in isotropic coordinates: ds^2 = -A^2 dt^2 + B^2 δij dx^i dx^j
    A = @. exp(ϕ)
    B = @. rsch / riso
    # B[1] is nan, because rsch = riso = 0 there
    # use a parabola to interpolate data point at riso = 0
    # this also ensures that B is symmetric arcoss riso = 0
    B2, B3, r2, r3 = B[2], B[3], riso[2], riso[3]
    B[1] = (r3^2*B2-r2^2*B3)/(r3^2-r2^2)

    (; Γ, K, ρc, M, riso, ρ, ϵ, p, A, B)
  end

  return data
end


function initialdata_equation!(::Val{:tov}, env, P::Project{:valencia1d}, prms)

  data = readdata_tov(prms["id_filename"])

  @unpack equation = P
  @unpack mesh = env
  @unpack eos = P.equation
  @assert data.K ≈ eos.K
  @assert data.Γ ≈ eos.Gamma

  equation.atm_density = data.ρc * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  @unpack rD, rSr, rτ   = get_dynamic_variables(mesh.cache)
  @unpack D, Sr, τ, ρ, p, ϵ, vr, r, max_v,
          init_D, init_Sr, init_τ, init_vr, init_p, init_ρ, init_ϵ,
          init_max_v    = get_static_variables(mesh.cache)
  @unpack α, βru, ∂αr, ∂βrrdu, sqrt_detγ,
          γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
          Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕrϕ, Γϕθϕ,
          Krr, Kθθ, Kϕϕ = get_static_variables(mesh.cache)

  rL, rR = extrema(r)
  ismirrored = -rL ≈ rR
  K, = mesh.tree.dims
  Npts = mesh.element.Npts
  isfv = mesh isa Mesh1d{<:FVElement}
  origin_duped = iseven(K) && !isfv
  origin_on_grid = (isodd(K) && isodd(Npts)) || (isfv && isodd(K))

  # need to split initial data into portion that lies inside and outside of star,
  # because derivatives of the hydro variables are discontinuous at the surface
  # although metric functions are supposed to be better conditioned there,
  # we also split them for the interpolation, because of potential contamination
  # with numerical error from patching
  rmax = data.riso[end]
  idx_origin = findfirst(ri -> ri >= 0.0, r)
  if ismirrored && origin_duped
    idx_origin += 1
    @assert r[idx_origin] ≈ 0.0
  end
  idx_exterior = findfirst(ri -> ri > rmax, r)
  has_exterior = !isnothing(idx_exterior)
  interior = has_exterior ? (idx_origin:idx_exterior-1) : (idx_origin:length(r))
  exterior = has_exterior ? (idx_exterior:length(mesh)) : nothing

  rint = view(r, interior)
  A, B, ∂Ar, ∂Br = [ similar(r) for _ in 1:4 ]

  ρ[interior] .= interpolate(rint, data.riso, data.ρ)
  ϵ[interior] .= interpolate(rint, data.riso, data.ϵ)
  p[interior] .= interpolate(rint, data.riso, data.p)
  A[interior] .= interpolate(rint, data.riso, data.A)
  B[interior] .= interpolate(rint, data.riso, data.B)

  if prms["perturbation_rho"]
    max_ρ = maximum(ρ)
    @. ρ += 1e-3 * max_ρ * randn(Float64)
  end

  if has_exterior
    rext = view(r, exterior)
    ρ[exterior] .= equation.atm_density
    p[exterior] .= @. eos(Pressure, Density, ρ[exterior])
    ϵ[exterior] .= @. eos(InternalEnergy, Density, ρ[exterior])
    # metric potentials of Schwarzschild solution in isotropic coordinates:
    #   ds^2 = -A^2 dt^2 + B^2 δij dx^i dx^j
    @unpack M = data
    A[exterior] .= @. (1-M/(2*rext))/(1+M/(2*rext))
    B[exterior] .= @. (1+M/(2*rext))^2
  end

  if ismirrored
    # fill other halves
    rev_interior = reverse(interior)
    if origin_duped
      mirror_interior = rev_interior.-length(rev_interior)
    elseif origin_on_grid
      TODO()
      mirror_interior = rev_interior[2:end].-length(rev_interior[2:end]).-1
      interior = interior[2:end]
      display(mirror_interior)
      display(interior)
      # error()
    else
      mirror_interior = rev_interior.-length(rev_interior)
    end
    for var in (ρ,ϵ,p,A,B)
      @views var[mirror_interior] .= var[interior]
    end
    if has_exterior
      mirror_exterior = length(exterior):-1:1
      for var in (ρ,ϵ,p,A,B)
        @views var[mirror_exterior] .= var[exterior]
      end
    end
  end

  # BasicInterpolators.jl cannot differentiate ...  so we diff numerically
  ∂Ar .= dg1d.differentiate(A, mesh)
  ∂Br .= dg1d.differentiate(B, mesh)

  θ = π/2

  @. α = A
  @. βru = 0.0
  @. ∂αr = ∂Ar
  @. ∂βrrdu = 0

  # γ_ij
  @. γrr = B^2
  @. γθθ = B^2*r^2
  @. γϕϕ = B^2*r^2*sin(θ)^2
  @. sqrt_detγ = B^3*r^2*sin(θ)

  # # γijuu = γ^ij
  # γrruu = @. 1/γrr
  # γθθuu = @. 1/γθθ
  # γϕϕuu = @. 1/γϕϕ

  # ∂iγjk := ∂γijk
  @. ∂γrrr = 2*B*∂Br
  @. ∂γrθθ = 2*(r*B^2+B*∂Br*r^2)
  @. ∂γrϕϕ = ∂γrθθ * sin(θ)^2
  @. ∂γθϕϕ = 0.0

  # Γijk := Γ^i_jk wrt γ_ij
  # TODO Verify these are correct
  @. Γrrr = ∂Br/B
  @. Γrθθ = (-r^2*B*∂Br-r*B^2)/B^2
  @. Γrϕϕ = Γrθθ*sin(θ)^2
  @. Γθrθ = (r^2*B*∂Br+r*B^2)/(r^2*B^2)
  @. Γθϕϕ = -sin(θ)*cos(θ)
  @. Γϕrϕ = (r^2*B*∂Br+r*B^2)/(r^2*B^2)
  @. Γϕθϕ = cos(θ)/sin(θ)

  # Kij := K_ij, γij is conformally flat
  @. Krr = 0.0
  @. Kθθ = 0.0
  @. Kϕϕ = 0.0

  @. vr = 0
  W = 1.0
  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D

  ref_D  = copy(D)
  ref_Sx = copy(Sx)
  ref_τ  = copy(τ)

  if P.prmsdb["GRHD"]["c2p_enforce_causal_atm"]
    broadcast_volume!(determine_atmosphere, equation, mesh)
  else
    @unpack c2p_freeze_atm = get_static_variables(mesh)
    c2p_freeze_atm .= 0.0 # atmosphere can always change
  end
  broadcast_volume!(cons2prim_valencia1d, equation, mesh)
  broadcast_volume!(maxspeed_valencia1d, equation, mesh)

  diff_D  = D.-ref_D
  diff_Sr = Sx.-ref_Sr
  diff_τ  = τ.-ref_τ
  absdiff_D  = maximum(abs, diff_D)
  absdiff_Sx = maximum(abs, diff_Sx)
  absdiff_τ  = maximum(abs, diff_τ)
  if absdiff_D > 1e-6*maximum(ref_D) ||
      absdiff_Sr > 1e-6*maximum(ref_Sr) ||
      absdiff_τ > 1e-6*maximum(ref_τ)
    @show absdiff_D, absdiff_Sr, absdiff_τ
    @show maximum(ref_D), maximum(ref_Sr), maximum(ref_τ)
    error("cons2prim altered initial data noticably, something is wrong")
  end

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v
  @. init_ρ     = ρ
  @. init_ϵ     = ϵ

end


function initialdata_equation!(::Val{:tov}, env, P::Project{:spherical1d}, prms)

  data = readdata_tov(prms["id_filename"])

  @unpack equation = P
  @unpack mesh = env
  @unpack eos = P.equation
  @assert data.K ≈ eos.K
  @assert data.Γ ≈ eos.Gamma

  equation.atm_density = data.ρc * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  @unpack D, Sr, τ = get_dynamic_variables(mesh.cache)
  @unpack ρ, p, ϵ, vr, r, max_v,
          init_D, init_Sr, init_τ, init_vr, init_p, init_ρ, init_ϵ,
          init_max_v = get_static_variables(mesh.cache)
  @unpack α, βru, ∂αr, ∂βrrdu,
          γrr, γθθ, γϕϕ, ∂γrrr, ∂γrθθ, ∂γrϕϕ, ∂γθϕϕ,
          Γrrr, Γrθθ, Γrϕϕ, Γθrθ, Γθϕϕ, Γϕrϕ, Γϕθϕ,
          Krr, Kθθ, Kϕϕ = get_static_variables(mesh.cache)

  rL, rR = extrema(r)
  ismirrored = -rL ≈ rR
  K, = mesh.tree.dims
  Npts = mesh.element.Npts
  isfv = mesh isa Mesh1d{<:FVElement}
  origin_duped = iseven(K) && !isfv
  origin_on_grid = (isodd(K) && isodd(Npts)) || (isfv && isodd(K))

  # need to split initial data into portion that lies inside and outside of star,
  # because derivatives of the hydro variables are discontinuous at the surface
  # although metric functions are supposed to be better conditioned there,
  # we also split them for the interpolation, because of potential contamination
  # with numerical error from patching
  rmax = data.riso[end]
  idx_origin = findfirst(ri -> ri >= 0.0, r)
  if ismirrored && origin_duped
    idx_origin += 1
    @assert r[idx_origin] ≈ 0.0
  end
  idx_exterior = findfirst(ri -> ri > rmax, r)
  has_exterior = !isnothing(idx_exterior)
  interior = has_exterior ? (idx_origin:idx_exterior-1) : (idx_origin:length(r))
  exterior = has_exterior ? (idx_exterior:length(mesh)) : nothing

  rint = view(r, interior)
  A, B, ∂Ar, ∂Br = [ similar(r) for _ in 1:4 ]

  ρ[interior] .= interpolate(rint, data.riso, data.ρ)
  ϵ[interior] .= interpolate(rint, data.riso, data.ϵ)
  p[interior] .= interpolate(rint, data.riso, data.p)
  A[interior] .= interpolate(rint, data.riso, data.A)
  B[interior] .= interpolate(rint, data.riso, data.B)

  if prms["perturbation_rho"]
    max_ρ = maximum(ρ)
    @. ρ += 1e-3 * max_ρ * randn(Float64)
  end

  if has_exterior
    rext = view(r, exterior)
    ρ[exterior] .= equation.atm_density
    p[exterior] .= @. eos(Pressure, Density, ρ[exterior])
    ϵ[exterior] .= @. eos(InternalEnergy, Density, ρ[exterior])
    # metric potentials of Schwarzschild solution in isotropic coordinates:
    #   ds^2 = -A^2 dt^2 + B^2 δij dx^i dx^j
    @unpack M = data
    A[exterior] .= @. (1-M/(2*rext))/(1+M/(2*rext))
    B[exterior] .= @. (1+M/(2*rext))^2
  end

  if ismirrored
    # fill other halves
    rev_interior = reverse(interior)
    if origin_duped
      mirror_interior = rev_interior.-length(rev_interior)
    elseif origin_on_grid
      TODO()
      mirror_interior = rev_interior[2:end].-length(rev_interior[2:end]).-1
      interior = interior[2:end]
      display(mirror_interior)
      display(interior)
      # error()
    else
      mirror_interior = rev_interior.-length(rev_interior)
    end
    for var in (ρ,ϵ,p,A,B)
      @views var[mirror_interior] .= var[interior]
    end
    if has_exterior
      mirror_exterior = length(exterior):-1:1
      for var in (ρ,ϵ,p,A,B)
        @views var[mirror_exterior] .= var[exterior]
      end
    end
  end

  # BasicInterpolators.jl cannot differentiate ...  so we diff numerically
  ∂Ar .= dg1d.differentiate(A, mesh)
  ∂Br .= dg1d.differentiate(B, mesh)

  θ = π/2

  # Watch out: 3+1 variables α,βⁱ,γᵢⱼ, Kᵢⱼ here belong to the physical metric
  # But Christoffels below belong to the reference metric!
  @. α = A
  @. βru = 0.0
  @. ∂αr = ∂Ar
  @. ∂βrrdu = 0

  @. γrr = B^2
  @. γθθ = B^2*r^2
  @. γϕϕ = B^2*r^2*sin(θ)^2

  @. ∂γrrr = 2*B*∂Br
  @. ∂γrθθ = 2*(r*B^2+B*∂Br*r^2)
  @. ∂γrϕϕ = ∂γrθθ * sin(θ)^2
  @. ∂γθϕϕ = 0.0

  @. Krr = 0.0
  @. Kθθ = 0.0
  @. Kϕϕ = 0.0

  # Christoffels of reference metric!
  @. Γrrr = 0.0
  @. Γrθθ = -r
  @. Γrϕϕ = -r*sin(θ)^2
  @. Γθrθ = 1.0/r
  @. Γθϕϕ = -sin(θ)*cos(θ)
  @. Γϕrϕ = 1.0/r
  @. Γϕθϕ = cot(θ)

  @. vr = 0
  W = 1.0
  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D

  ref_D  = copy(D)
  ref_Sr = copy(Sr)
  ref_τ  = copy(τ)

  if P.prmsdb["GRHD"]["c2p_enforce_causal_atm"]
    broadcast_volume!(determine_atmosphere, equation, mesh)
  else
    @unpack c2p_freeze_atm = get_static_variables(mesh)
    c2p_freeze_atm .= 0.0 # atmosphere can always change
  end
  broadcast_volume!(cons2prim_spherical1d, equation, mesh)
  broadcast_volume!(maxspeed_spherical1d, equation, mesh)

  diff_D  = D.-ref_D
  diff_Sr = Sr.-ref_Sr
  diff_τ  = τ.-ref_τ
  absdiff_D  = maximum(abs, diff_D)
  absdiff_Sr = maximum(abs, diff_Sr)
  absdiff_τ  = maximum(abs, diff_τ)
  if absdiff_D > 1e-6*maximum(ref_D) ||
      absdiff_Sr > 1e-6*maximum(ref_Sr) ||
      absdiff_τ > 1e-6*maximum(ref_τ)
    @show absdiff_D, absdiff_Sr, absdiff_τ
    @show maximum(ref_D), maximum(ref_Sr), maximum(ref_τ)
    # error("cons2prim altered initial data noticably, something is wrong")
  end

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v
  @. init_ρ     = ρ
  @. init_ϵ     = ϵ

  Mtot = dg1d.integrate(mesh) do i
    v = D[i] * sqrt(γrr[i]*γθθ[i]*γϕϕ[i]) * 4*π
    v
  end
  if dg1d.is_domain_symmetric(mesh)
    Mtot /= 2
  end
  Madm = data.M
  @show Mtot, Madm

end


function initialdata_equation!(::Val{:tov}, env, P::Project{:rescaled_spherical1d}, prms)

  data = readdata_tov(prms["id_filename"])

  @unpack equation = P
  @unpack mesh = env
  @unpack eos = P.equation
  @assert data.K ≈ eos.K
  @assert data.Γ ≈ eos.Gamma

  equation.atm_density = data.ρc * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  @unpack rD, rSr, rτ = get_dynamic_variables(mesh.cache)
  @unpack ρ, p, ϵ, vr, r, max_v,
          D, Sr, τ,
          init_D, init_Sr, init_τ, init_ρ, init_ϵ,
          init_vr, init_p, init_max_v = get_static_variables(mesh.cache)
  @unpack A, ∂Ar, B, ∂Br, γrr = get_static_variables(mesh.cache)

  rL, rR = extrema(r)
  ismirrored = -rL ≈ rR
  K, = mesh.tree.dims
  Npts = mesh.element.Npts
  isfv = mesh isa Mesh1d{<:FVElement}
  origin_duped = iseven(K) && !isfv
  origin_on_grid = (isodd(K) && isodd(Npts)) || (isfv && isodd(K))

  # need to split initial data into portion that lies inside and outside of star,
  # because derivatives of the hydro variables are discontinuous at the surface
  # although metric functions are supposed to be better conditioned there,
  # we also split them for the interpolation, because of potential contamination
  # with numerical error from patching
  rmax = data.riso[end]
  idx_origin = findfirst(ri -> ri >= 0.0, r)
  if ismirrored && origin_duped
    idx_origin += 1
    @assert r[idx_origin] ≈ 0.0
  end
  idx_exterior = findfirst(ri -> ri > rmax, r)
  has_exterior = !isnothing(idx_exterior)
  interior = has_exterior ? (idx_origin:idx_exterior-1) : (idx_origin:length(r))
  exterior = has_exterior ? (idx_exterior:length(mesh)) : nothing

  rint = view(r, interior)

  ρ[interior] .= interpolate(rint, data.riso, data.ρ)
  ϵ[interior] .= interpolate(rint, data.riso, data.ϵ)
  p[interior] .= interpolate(rint, data.riso, data.p)
  A[interior] .= interpolate(rint, data.riso, data.A)
  B[interior] .= interpolate(rint, data.riso, data.B)

  if prms["perturbation_rho"]
    max_ρ = maximum(ρ)
    @. ρ += 1e-3 * max_ρ * randn(Float64)
  end

  if has_exterior
    rext = view(r, exterior)
    ρ[exterior] .= equation.atm_density
    p[exterior] .= @. eos(Pressure, Density, ρ[exterior])
    ϵ[exterior] .= @. eos(InternalEnergy, Density, ρ[exterior])
    # metric potentials of Schwarzschild solution in isotropic coordinates:
    #   ds^2 = -A^2 dt^2 + B^2 δij dx^i dx^j
    @unpack M = data
    A[exterior] .= @. (1-M/(2*rext))/(1+M/(2*rext))
    B[exterior] .= @. (1+M/(2*rext))^2
  end

  if ismirrored
    # fill other halves
    rev_interior = reverse(interior)
    if origin_duped
      mirror_interior = rev_interior.-length(rev_interior)
    elseif origin_on_grid
      TODO()
      mirror_interior = rev_interior[2:end].-length(rev_interior[2:end]).-1
      interior = interior[2:end]
      display(mirror_interior)
      display(interior)
      # error()
    else
      mirror_interior = rev_interior.-length(rev_interior)
    end
    for var in (ρ,ϵ,p,A,B)
      @views var[mirror_interior] .= var[interior]
    end
    if has_exterior
      mirror_exterior = length(exterior):-1:1
      for var in (ρ,ϵ,p,A,B)
        @views var[mirror_exterior] .= var[exterior]
      end
    end
  end

  # BasicInterpolators.jl cannot differentiate ...  so we diff numerically
  ∂Ar .= dg1d.differentiate(A, mesh)
  ∂Br .= dg1d.differentiate(B, mesh)
  @. γrr = B^2

  @. vr = 0
  W = 1.0
  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sr = W^2 * ρ * h * vr
  @. τ  = W^2 * ρ * h - p - D
  @. rD  = r*B^3*D
  @. rSr = r*B^3*Sr
  @. rτ  = r*B^3*τ

  ref_D  = copy(D)
  ref_Sr = copy(Sr)
  ref_τ  = copy(τ)

  broadcast_volume!(unscale_conservatives, equation, mesh)
  impose_symmetry!(P, mesh)
  if P.prmsdb["GRHD"]["c2p_enforce_causal_atm"]
    broadcast_volume!(determine_atmosphere, equation, mesh)
  else
    @unpack c2p_freeze_atm = get_static_variables(mesh)
    c2p_freeze_atm .= 0.0 # atmosphere can always change
  end
  broadcast_volume!(cons2prim_rescaled_spherical1d, equation, mesh)
  broadcast_volume!(maxspeed_rescaled_spherical1d, equation, mesh)

  diff_D  = D.-ref_D
  diff_Sr = Sr.-ref_Sr
  diff_τ  = τ.-ref_τ
  absdiff_D  = maximum(abs, diff_D)
  absdiff_Sr = maximum(abs, diff_Sr)
  absdiff_τ  = maximum(abs, diff_τ)
  if absdiff_D > 1e-6*maximum(ref_D) ||
      absdiff_Sr > 1e-6*maximum(ref_Sr) ||
      absdiff_τ > 1e-6*maximum(ref_τ)
    @show absdiff_D, absdiff_Sr, absdiff_τ
    @show maximum(ref_D), maximum(ref_Sr), maximum(ref_τ)
    error("cons2prim altered initial data noticably, something is wrong")
  end

  @. init_D     = D
  @. init_Sr    = Sr
  @. init_τ     = τ
  @. init_vr    = vr
  @. init_p     = p
  @. init_max_v = max_v
  @. init_ρ     = ρ
  @. init_ϵ     = ϵ

end


function initialdata_equation!(::Val{:tov}, env, P::Project{:doublecartoon}, prms)
  data = readdata_tov(prms["id_filename"])

  @unpack equation = P
  @unpack mesh = env
  @unpack eos = P.equation
  @assert data.K ≈ eos.K
  @assert data.Γ ≈ eos.Gamma

  equation.atm_density = data.ρc * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  @unpack rD, rSx, rτ = get_dynamic_variables(mesh.cache)
  @unpack ρ, p, ϵ, vx, r, max_v,
          D, Sx, τ = get_static_variables(mesh.cache)
  @unpack α, βux, βuy, βuz,
          ∂αx, ∂αy, ∂αz,
          ∂βuxx, ∂βuyx, ∂βuzx,
          ∂βuxy, ∂βuyy, ∂βuzy,
          ∂βuxz, ∂βuyz, ∂βuzz,
          ψ4, ∂ψ4x,
          γxx, γxy, γxz,
          γyy, γyz,
          γzz,
          sqrtdetγ,
          ∂γxxx, ∂γxxy, ∂γxxz,
          ∂γxyy, ∂γxyz,
          ∂γxzz,
          ∂γyxx, ∂γyxy, ∂γyxz,
          ∂γyyy, ∂γyyz,
          ∂γyzz,
          ∂γzxx, ∂γzxy, ∂γzxz,
          ∂γzyy, ∂γzyz,
          ∂γzzz = get_static_variables(mesh.cache)

  rL, rR = extrema(r)
  ismirrored = -rL ≈ rR
  K, = mesh.tree.dims
  Npts = mesh.element.Npts
  isfv = mesh isa Mesh1d{<:FVElement}
  origin_duped = iseven(K) && !isfv
  origin_on_grid = (isodd(K) && isodd(Npts)) || (isfv && isodd(K))

  # need to split initial data into portion that lies inside and outside of star,
  # because derivatives of the hydro variables are discontinuous at the surface
  # although metric functions are supposed to be better conditioned there,
  # we also split them for the interpolation, because of potential contamination
  # with numerical error from patching
  rmax = data.riso[end]
  idx_origin = findfirst(ri -> ri >= 0.0, r)
  if ismirrored && origin_duped
    idx_origin += 1
    @assert r[idx_origin] ≈ 0.0
  end
  idx_exterior = findfirst(ri -> ri > rmax, r)
  has_exterior = !isnothing(idx_exterior)
  interior = has_exterior ? (idx_origin:idx_exterior-1) : (idx_origin:length(r))
  exterior = has_exterior ? (idx_exterior:length(mesh)) : nothing

  rint = view(r, interior)

  ρ[interior]  .= interpolate(rint, data.riso, data.ρ)
  ϵ[interior]  .= interpolate(rint, data.riso, data.ϵ)
  p[interior]  .= interpolate(rint, data.riso, data.p)
  α[interior]  .= interpolate(rint, data.riso, data.A)
  ψ4[interior] .= interpolate(rint, data.riso, data.B).^2

  if prms["perturbation_rho"]
    max_ρ = maximum(ρ)
    @. ρ += 1e-3 * max_ρ * randn(Float64)
  end

  if has_exterior
    rext = view(r, exterior)
    ρ[exterior] .= equation.atm_density
    p[exterior] .= @. eos(Pressure, Density, ρ[exterior])
    ϵ[exterior] .= @. eos(InternalEnergy, Density, ρ[exterior])
    # metric potentials of Schwarzschild solution in isotropic coordinates:
    #   ds^2 = -A^2 dt^2 + B^2 δij dx^i dx^j
    @unpack M = data
    α[exterior]  .= @. (1-M/(2*rext))/(1+M/(2*rext))
    ψ4[exterior] .= @. (1+M/(2*rext))^4
  end

  if ismirrored
    # fill other halves
    rev_interior = reverse(interior)
    if origin_duped
      mirror_interior = rev_interior.-length(rev_interior)
    elseif origin_on_grid
      TODO()
      mirror_interior = rev_interior[2:end].-length(rev_interior[2:end]).-1
      interior = interior[2:end]
      display(mirror_interior)
      display(interior)
      # error()
    else
      mirror_interior = rev_interior.-length(rev_interior)
    end
    for var in (ρ,ϵ,p,α,ψ4)
      @views var[mirror_interior] .= var[interior]
    end
    if has_exterior
      mirror_exterior = length(exterior):-1:1
      for var in (ρ,ϵ,p,α,ψ4)
        @views var[mirror_exterior] .= var[exterior]
      end
    end
  end

  # BasicInterpolators.jl cannot differentiate ...  so we diff numerically
  ∂αx .= dg1d.differentiate(α, mesh)
  ∂ψ4x .= dg1d.differentiate(ψ4, mesh)

  # only relevant components in spherical symmetry and when evaluated at y=z=0
  @. γxx = γyy = γzz = ψ4
  @. ∂γxxx = ∂γxyy = ∂γxzz = ∂ψ4x

  detγ = @. 2*γyz*γxy*γxz + γzz*γyy*γxx - (γyy*γxz^2 + γyz^2*γxx + γzz*γxy^2)
  @. sqrtdetγ = sqrt(detγ)

  @. vx = 0 # = vy = vz
  W = 1.0
  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sx = W^2 * ρ * h * vx
  @. τ  = W^2 * ρ * h - p - D

  @. rD  = sqrtdetγ*D
  @. rSx = sqrtdetγ*Sx
  @. rτ  = sqrtdetγ*τ

  ref_D  = copy(D)
  ref_Sx = copy(Sx)
  ref_τ  = copy(τ)

  broadcast_volume!(unscale_conservatives_doublecartoon, equation, mesh)
  if P.prmsdb["GRHD"]["c2p_enforce_causal_atm"]
    broadcast_volume!(determine_atmosphere, equation, mesh)
  else
    @unpack c2p_freeze_atm = get_static_variables(mesh)
    c2p_freeze_atm .= 0.0 # atmosphere can always change
  end
  broadcast_volume!(cons2prim_doublecartoon, equation, mesh)
  broadcast_volume!(maxspeed_doublecartoon, equation, mesh)

  diff_D  = D.-ref_D
  diff_Sx = Sx.-ref_Sx
  diff_τ  = τ.-ref_τ
  absdiff_D  = maximum(abs, diff_D)
  absdiff_Sx = maximum(abs, diff_Sx)
  absdiff_τ  = maximum(abs, diff_τ)
  if absdiff_D > 1e-6*maximum(ref_D) ||
      absdiff_Sx > 1e-6*maximum(ref_Sx) ||
      absdiff_τ > 1e-6*maximum(ref_τ)
    @show absdiff_D, absdiff_Sx, absdiff_τ
    @show maximum(ref_D), maximum(ref_Sx), maximum(ref_τ)
    error("cons2prim altered initial data noticably, something is wrong")
  end

end


function initialdata_equation!(::Val{:tov}, env, P::Project{:cartoon}, prms)
  data = readdata_tov(prms["id_filename"])

  @unpack equation = P
  @unpack mesh = env
  @unpack eos = P.equation
  @assert data.K ≈ eos.K
  @assert data.Γ ≈ eos.Gamma

  equation.atm_density = data.ρc * equation.atm_factor
  equation.atm_factor = prms["atm_factor"]
  equation.atm_threshold = prms["atm_threshold_factor"]

  @unpack rD, rSx, rSz, rτ = get_dynamic_variables(mesh.cache)
  @unpack ρ, p, ϵ, vx, vz, r,
          D, Sx, Sz, τ = get_static_variables(mesh.cache)
  @unpack α, βux, βuy, βuz,
          ∂αx, ∂αy, ∂αz,
          ∂βuxx, ∂βuyx, ∂βuzx,
          ∂βuxy, ∂βuyy, ∂βuzy,
          ∂βuxz, ∂βuyz, ∂βuzz,
          ψ4, ∂ψ4x, ∂ψ4z,
          γxx, γxy, γxz,
          γyy, γyz,
          γzz,
          sqrtdetγ,
          ∂γxxx, ∂γxxy, ∂γxxz,
          ∂γxyy, ∂γxyz,
          ∂γxzz,
          ∂γyxx, ∂γyxy, ∂γyxz,
          ∂γyyy, ∂γyyz,
          ∂γyzz,
          ∂γzxx, ∂γzxy, ∂γzxz,
          ∂γzyy, ∂γzyz,
          ∂γzzz = get_static_variables(mesh.cache)

  rL, rR = extrema(r)
  ismirrored = -rL ≈ rR
  K, = mesh.tree.dims
  Npts = mesh.element.Npts
  isfv = mesh isa Mesh1d{<:FVElement}
  origin_duped = iseven(K) && !isfv
  origin_on_grid = (isodd(K) && isodd(Npts)) || (isfv && isodd(K))

  # need to split initial data into portion that lies inside and outside of star,
  # because derivatives of the hydro variables are discontinuous at the surface
  # although metric functions are supposed to be better conditioned there,
  # we also split them for the interpolation, because of potential contamination
  # with numerical error from patching
  rmax = data.riso[end]
  idx_origin = findfirst(ri -> ri >= 0.0, r)
  if ismirrored && origin_duped
    idx_origin += 1
    @assert r[idx_origin] ≈ 0.0
  end
  exterior = r .> rmax
  interior = .!exterior
  has_exterior = any(exterior)

  rint = view(r, interior)

  ρ[interior]  .= interpolate(rint, data.riso, data.ρ)
  ϵ[interior]  .= interpolate(rint, data.riso, data.ϵ)
  p[interior]  .= interpolate(rint, data.riso, data.p)
  α[interior]  .= interpolate(rint, data.riso, data.A)
  ψ4[interior] .= interpolate(rint, data.riso, data.B).^2

  if prms["perturbation_rho"]
    max_ρ = maximum(ρ)
    @. ρ += 1e-3 * max_ρ * randn(Float64)
  end

  if has_exterior
    rext = view(r, exterior)
    ρ[exterior] .= equation.atm_density
    p[exterior] .= @. eos(Pressure, Density, ρ[exterior])
    ϵ[exterior] .= @. eos(InternalEnergy, Density, ρ[exterior])
    # metric potentials of Schwarzschild solution in isotropic coordinates:
    #   ds^2 = -A^2 dt^2 + B^2 δij dx^i dx^j
    @unpack M = data
    α[exterior]  .= @. (1-M/(2*rext))/(1+M/(2*rext))
    ψ4[exterior] .= @. (1+M/(2*rext))^4
  end

  if ismirrored
    # fill other halves
    rev_interior = reverse(interior)
    if origin_duped
      mirror_interior = rev_interior.-length(rev_interior)
    elseif origin_on_grid
      TODO()
      mirror_interior = rev_interior[2:end].-length(rev_interior[2:end]).-1
      interior = interior[2:end]
      display(mirror_interior)
      display(interior)
      # error()
    else
      mirror_interior = rev_interior.-length(rev_interior)
    end
    for var in (ρ,ϵ,p,α,ψ4)
      @views var[mirror_interior] .= var[interior]
    end
    if has_exterior
      mirror_exterior = length(exterior):-1:1
      for var in (ρ,ϵ,p,α,ψ4)
        @views var[mirror_exterior] .= var[exterior]
      end
    end
  end

  # BasicInterpolators.jl cannot differentiate ...  so we diff numerically
  ∂αx .= dg1d.differentiate(α, mesh, dir=:x)
  ∂αz .= dg1d.differentiate(α, mesh, dir=:y)
  ∂ψ4x .= dg1d.differentiate(ψ4, mesh, dir=:x)
  ∂ψ4z .= dg1d.differentiate(ψ4, mesh, dir=:y) # logical coordinate y = physical coordinate z

  # only relevant components in spherical symmetry and when evaluated at y=z=0
  @. γxx = γyy = γzz = ψ4
  @. ∂γxxx = ∂γxyy = ∂γxzz = ∂ψ4x
  @. ∂γzxx = ∂γzyy = ∂γzzz = ∂ψ4z

  detγ = @. 2*γyz*γxy*γxz + γzz*γyy*γxx - (γyy*γxz^2 + γyz^2*γxx + γzz*γxy^2)
  @. sqrtdetγ = sqrt(detγ)

  @. vx = vz = 0 # = vy
  W = 1.0
  h = @. 1 + ϵ + p/ρ
  @. D  = W * ρ
  @. Sx = W^2 * ρ * h * vx
  @. Sz = W^2 * ρ * h * vz
  @. τ  = W^2 * ρ * h - p - D

  @. rD  = sqrtdetγ*D
  @. rSx = sqrtdetγ*Sx
  @. rSz = sqrtdetγ*Sz
  @. rτ  = sqrtdetγ*τ

  ref_D  = copy(D)
  ref_Sx = copy(Sx)
  ref_τ  = copy(τ)

  broadcast_volume!(unscale_conservatives_cartoon, equation, mesh)
  if P.prmsdb["GRHD"]["c2p_enforce_causal_atm"]
    broadcast_volume!(determine_atmosphere, equation, mesh)
  else
    @unpack c2p_freeze_atm = get_static_variables(mesh)
    c2p_freeze_atm .= 0.0 # atmosphere can always change
  end
  broadcast_volume!(cons2prim_cartoon, equation, mesh)
  broadcast_volume!(maxspeed_cartoon, equation, mesh)

  diff_D  = D.-ref_D
  diff_Sx = Sx.-ref_Sx
  diff_τ  = τ.-ref_τ
  absdiff_D  = maximum(abs, diff_D)
  absdiff_Sx = maximum(abs, diff_Sx)
  absdiff_τ  = maximum(abs, diff_τ)
  if absdiff_D > 1e-6*maximum(ref_D) ||
      absdiff_Sx > 1e-6*maximum(ref_Sx) ||
      absdiff_τ > 1e-6*maximum(ref_τ)
    @show absdiff_D, absdiff_Sx, absdiff_τ
    @show maximum(ref_D), maximum(ref_Sx), maximum(ref_τ)
    error("cons2prim altered initial data noticably, something is wrong")
  end

end


function interpolate(x, x_ref, y_ref)
  intrp = CubicSplineInterpolator(x_ref, y_ref, WeakBoundaries())
  return intrp.(x)
end


function interpolate_derivative(x, x_ref, y_ref)
  intrp = CubicSplineInterpolator(x_ref, y_ref, WeakBoundaries())
  return intrp.(x)
end


#######################################################################
#                    Initialdata entropy variables                    #
#######################################################################


function initialdata_entropy_variables!(::Val, P, prms)
  TODO(initialdata_entropy_variables!)
end

#######################################################################
#                          InitialData HRSC                           #
#######################################################################

initialdata_hrsc!(kind, env, P, hrsc::Nothing, prms) = nothing
function initialdata_hrsc!(kind, env, P, hrsc, prms)
  TODO(initialdata_hrsc!)
end

initialdata_hrsc!(kind, env, P::Project, hrsc::HRSC.EntropyViscosity, prms) = TODO(hrsc)

function initialdata_hrsc!(kind, env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, prms)
  @unpack mu, tmp_mu, mu_m1      = get_cell_variables(env.cache)
  mda_prms = parameters(:HRSC)
  mda_prms["method"] = "av"
  mda_prms["av_method"] = "mda"
  mda_prms["mda_cmax"] = 1.0
  mda_prms["mda_n_high"] = 4.0
  mda_prms["mda_n_low"] = mda_prms["mda_n_high"]*0.99
  dg1d.check_parameters(:HRSC, mda_prms)
  mda_hrsc = HRSC.make_HRSC(env.mesh, mda_prms)
  @unpack cellmax_v = get_cell_variables(env.cache)
  @unpack max_v = get_static_variables(env.cache)
  for (k,vk) in enumerate(eachcell(env.mesh, max_v))
    cellmax_v[k] = maximum(abs, vk)
  end
  if formulation(P) === :valencia1d || formulation(P) === :doublecartoon
    @unpack rD = get_dynamic_variables(env.cache)
    HRSC.compute_viscosity!(
        mu, tmp_mu,
        (log10.(abs.(rD)),), cellmax_v,
        mda_hrsc.av)
  else
    @unpack D = get_dynamic_variables(env.cache)
    HRSC.compute_viscosity!(
        mu, tmp_mu,
        (log10.(abs.(D)),), cellmax_v,
        mda_hrsc.av)
  end
  @. mu_m1 = mu
end

function initialdata_hrsc!(kind, env, P::Project, hrsc::HRSC.SmoothedArtificialViscosity, prms)
  @unpack mesh = env
  @unpack mu, mu_m1 = get_cell_variables(mesh)
  @unpack smoothed_mu = get_static_variables(mesh)
  initialdata_hrsc!(kind, env, P, hrsc.av, prms)
  hrsc.smoother(smoothed_mu, mu, env.mesh, mesh.tree.periodic[1])
end

initialdata_hrsc!(kind, env, P::Project, hrsc::HRSC.AbstractReconstruction, prms) = nothing

#######################################################################
#                           InitialData TCI                           #
#######################################################################

function initialdata_tci!(::Val, P, tci, prms)
  TODO(initialdata_tci!)
end

#######################################################################
#                              Smoothing                              #
#######################################################################

function initialdata_smooth!(env, P::Project, prms)

  @unpack mesh = env
  @unpack D = get_dynamic_variables(mesh)
  K, = mesh.tree.dims
  flag = zeros(Float64, K)

  tci_prms = dg1d.parameters(:TCI)
  tci_prms["method"] = "mda"
  tci_prms["variables"] = [ :D ]
  dg1d.check_parameters(:TCI, tci_prms)
  tci = TCI.make_TCI(mesh, tci_prms)
  TCI.compute_indicator!(flag, log10.(abs.(D)), tci)

  tmp_flag = copy(flag)
  for k = 1:K
    k_lhs, k_rhs = periodic_index(k-1, K), periodic_index(k+1, K)
    if !isperiodic(mesh) && k == 1
      tmp_flag[k] = (flag[k] + flag[k_rhs]) / 2
    elseif !isperiodic(mesh) && k == K
      tmp_flag[k] = (flag[k_lhs] + flag[k]) / 2
    else
      tmp_flag[k] = (flag[k_lhs] + flag[k] + flag[k_rhs]) / 3
    end
  end
  @. flag = tmp_flag

  @unpack D, Sr, τ = get_dynamic_variables(mesh)
  hrsc_prms = parameters(:HRSC)
  hrsc_prms["method"] = "filter"
  hrsc_prms["filter_method"] = "bernstein"
  dg1d.check_parameters(:HRSC, hrsc_prms)
  bernstein_hrsc = HRSC.make_HRSC(mesh, hrsc_prms)

  HRSC.reconstruct!(D,  flag, bernstein_hrsc, isperiodic=false)
  HRSC.reconstruct!(Sr, flag, bernstein_hrsc, isperiodic=false)
  HRSC.reconstruct!(τ,  flag, bernstein_hrsc, isperiodic=false)

  broadcast_volume!(cons2prim, P.equation, mesh)
  broadcast_volume!(maxspeed, P.equation, mesh)
  impose_symmetry_sources!(P, mesh)

end
