const TO = TimerOutput()


function get_testparfiles()
  path = normpath(joinpath(@__DIR__, "..", "test", "IntegrationTests", "refs", "testconfig.toml"))
  if !isfile(path)
    error("failed to locate testconfig $path")
  end
  config = TOML.parsefile(path)
  list = [ k for (k,v) in pairs(config) if v isa Dict ]
  return sort(list)
end

function testparfiles()
  list = get_testparfiles()
  println("Available list of test parameter files:")
  pretty_table(hcat(1:length(list), list), alignment=[:r,:l], noheader=true)
  println("Use with `main(\"@test/<name>\")`")
end


function normalize_parfile(path::AbstractString)
  filename = if isdir(path)
    name = basename(path)
    fname = joinpath(path, string(name, ".toml"))
  else
    path
  end
  if !isfile(filename)
    error("Cannot locate a parameter file at $path")
  end
  return filename
end
function normalize_parfile(nr::Integer)
  testpars = get_testparfiles()
  if !(1 <= nr <= length(testpars))
    error("There is no test parfile with nr $nr, see testparfiles() for available parfiles")
  end
  parfile = testpars[nr]
  path = normpath(joinpath(@__DIR__, "..", "test", "IntegrationTests", "refs", parfile, "$parfile.toml"))
  return path
end


"""
    main(parfile)

Run program specified by a `parfile` in TOML format.
"""
function the_main(parfile::String, parameter_overrides::Pair{String,<:Any}...)

  parfile = abspath(parfile)
  prms = TOML.parsefile(parfile)
  flat_prms = flatten(prms)
  for (k,v) in parameter_overrides
    flat_prms[k] = v
  end
  prms = unflatten(flat_prms)

  dir = dirname(parfile)
  if basename(dir) == splitext(basename(parfile))[1]
    # we are running parameter file contained in an output folder
    # so we backup that folder and run the parameter file one directory above
    parfile = string(dir, ".toml")
  end
  # backup any previous output dir and set up a fresh one
  outputdir = make_outputdir(parfile)
  parfile = joinpath(outputdir, basename(parfile))
  open(parfile, "w") do file
    TOML.print(file, prms)
  end
  cp(joinpath(@__DIR__,"..","Manifest.toml"), joinpath(outputdir,"Manifest.toml"))
  open(joinpath(outputdir, "dirty"), "w") do file
    println(file, "git commit hash: $(githash())")
    println(file, '='^100)
    println(file, gitdirty())
  end

  env, project, bdryconds, prms = setup_env(parfile)
  rhs_fn, timestep_fn = setup_callbacks(env, project, bdryconds, prms, outputdir)
  # TODO Use PrettyTables.jl for summary
  # TODO Dump versioninfo() to some file
  # TODO Dump Pkg.status() to some file
  # TODO Dump gethostname() and other system info
  # TODO summarize current setup once and log to screen
  # println(summarize(env.mesh))
  # println(summarize(env.cache))
  # println(summarize(env.callbacks))
  # println(summarize(env.callbacks_substep))
  # println(summarize(project))

  display(prms)
  # dump database summary to output folder
  open(joinpath(outputdir, "parameters.log"), "w") do f
    showparameters(f, prms, annotate=true)
  end
  # dump a parfile with current settings (including defaults)
  open(joinpath(outputdir, "all_parameters.toml"), "w") do f
    TOML.print(f, Dict(prms))
  end

  reset_timer!(TO)
  @timeit TO "Evolution" begin
    # Let's go!!!
    evolve!(env, rhs_fn, timestep_fn, prms)
  end

  # finish timer, progress bar
  timercb = env.callbacks.callbacks[end-1]
  timercb([], 0.0, 0, force=true)
  progresscb = env.callbacks.callbacks[end-1]
  progresscb([], 0.0, 0, force=true)

  return
end

main(p::Union{Integer,AbstractString}, parameter_overrides::Pair{String,<:Any}...) =
  the_main(normalize_parfile(p), parameter_overrides...)

load_parameters(parfile::AbstractString) = parse_parameters(TOML.parsefile(parfile))

function parse_parameters(user_prms::AbstractDict)

  # switch to a flat dict where keys of nested dicts are concatenated with dots
  user_ks, user_vals = keys(user_prms), values(user_prms)
  user_vals = flatten.(user_vals)
  for sub in user_vals
    for (k,v) in pairs(sub)
      sub[k] = Sandbox.parse_expression_string(v)
    end
  end
  user_prms = Dict(collect(user_ks) .=> user_vals)

  ### sanitize parameter sections
  # supported top-level headers that aren't user projects
  default_projects = [ :Mesh, :Evolution, :Output, :TCI, :HRSC, :EquationOfState, :Log ]

  for (k,v) in user_prms
    if !(v isa AbstractDict)
      error("Invalid parameter file: No top-level parameters allowed, found '$k = $v'")
    end
  end

  # find a parameter section for which somebody registered a project with that name
  all_projects = Symbol.(keys(user_prms))
  maybe_user_projects = filter(all_projects) do k
    return !(k in default_projects)
  end
  registered_projects = filter(isregistered ∘ Symbol, maybe_user_projects)
  not_registered_projects = setdiff(maybe_user_projects, registered_projects)
  if length(not_registered_projects) > 0
    error("Invalid parameter file: Found unknown top-level sections: $(not_registered_projects)")
  end
  if length(registered_projects) == 0
    error("Invalid parameter file: Missing a parameter section for a project, don't know which project to run ...")
  end
  if length(registered_projects) > 1
    error("Invalid parameter file: Only one project parameter section allowed, found multiple: $(registered_projects)")
  end

  # we found a user project
  project_name = Symbol(first(registered_projects))

  # load default parameters
  default_subs = []
  for s in default_projects
    defaults = parameters(Val(s))
    for (k,v) in pairs(defaults)
      defaults[k] = Sandbox.parse_expression_string(v)
    end
    push!(default_subs, defaults)
  end
  prms = Dict(string.(default_projects) .=> default_subs)
  prms[string(project_name)] = parameters(project_name)
  # TODO parameters(Val) does not guarantee that dicts returned from parameter are flat.
  prms = Dict(k => flatten(v) for (k,v) in pairs(prms))

  # merge defaults with with user parameters
  for (sec, u_prms) in pairs(user_prms)
    def_prms = prms[sec]
    for (k,v) in pairs(u_prms)
      if !haskey(def_prms, k)
        @warn "Unsupported parameter in section [ $sec ]: '$k = $v'."
      end
      def_prms[k] = v
    end
  end

  # use a database to keep track of which parameters will be used
  prms = ParameterDB( Dict( k => ParameterDict(v) for (k,v) in pairs(prms) ) )

  return project_name, prms
end


function setup_env(parfile::Union{AbstractString,Integer})
  parfile = normalize_parfile(parfile)
  user_prms = TOML.parsefile(parfile)
  project_name, prms = parse_parameters(user_prms)
  return setup_env(project_name, prms)
end


function setup_env(project_name, prms)
  mprms = prms["Mesh"]
  mesh = Mesh(; N=mprms["n"], K=mprms["k"], range=mprms["range"], periodic=mprms["periodic"],
              scheme=mprms["scheme"], basis=mprms["basis"], dg_kind=mprms["dg_kind"],
              filter_parameters=Tuple(mprms["filter_parameters"]))
  callbacks = CallbackSet()
  callbacks_substeps = CallbackSet()
  env = Environment(mesh, mesh.cache, callbacks, callbacks_substeps)
  # Setup the project. This involves
  # - registering variables in cache
  # - setting up initial data
  # - setting up boundary conditions
  # - setting up callbacks
  # Need to use Val(project_name) for dispatch here, because project_name is a runtime variable.
  Project = project_type(Val(project_name))
  # TODO Till I sorted out the issue with where to place bdry condition types, I have to
  # return them here and forward them below
  project, bdryconds = Project(env, prms)
  return env, project, bdryconds, prms
end


function setup_callbacks(env, project, bdryconds, prms, outputdir)

  # wrap project interface methods for useage with time stepper
  function wrapped_rhs!(state_du, state_u, t)
    @timeit TO "Wrap dynamic" begin
      wrap_dynamic_variables!(env.cache, state_u)
    end
    @timeit TO "Wrap rhs" begin
      wrap_rhs_variables!(env.cache, state_du)
    end
    # provided by project interface
    @timeit TO "rhs" begin
      rhs!(env, project, bdryconds)
    end
  end
  CFL = prms["Evolution"]["cfl"]
  function wrapped_timestep!(state_u, t, n)
    @timeit TO "Wrap dynamic" begin
      wrap_dynamic_variables!(env.cache, state_u)
    end
    # provided by project interface
    @timeit TO "timestep" begin
      # need to apply CFL for all projects here,
      # because different places might call this function (e.g. SaveCallback, Evolution)
      # and compute a timestep multiple times, although, that should be deprecated in the future
      CFL * timestep!(env, project)
    end
  end
  function wrap_dynamic(state_u, t)
    @timeit TO "Wrap dynamic" begin
      wrap_dynamic_variables!(env.cache, state_u)
    end
  end
  if prms["Mesh"]["filter_enable"]
    register_variables!(env.mesh.cache, static_variablenames=(:buf_spectral_filter,))
  end
  function cb_spectral_filter!(state_u, t)
    @unpack buf_spectral_filter = get_static_variables(env.mesh)
    for var in values(env.cache.dynamic_variables.datadict)
      spectral_filter!(env.mesh, var, buf_spectral_filter)
    end
  end

  # Setup project independent callbacks
  # For the time being, we let the Projects freely push callbacks into env.callbacks.
  # This should make no difference as long as we 1) don't run in parallel and 2) don't allow
  # multiple projects. Might want to move that logic later from Project to here.
  # TODO Add an AliveCallback that manages a 'status' file in the output folder
  # TODO Add priority ordering to CallbackSet and callbacks.
  # TODO Drop unused callbacks from callbackset upon creation

  mesh = env.mesh
  # Add some 'global' callbacks
  savecb_timing = CallbackTiming(every_dt=prms["Output"]["every_dt"],
                                 every_iteration=prms["Output"]["every_iteration"])
  vars0d = prms["Output"]["variables0d"]
  savecb_0d = SaveCallback(Symbol.(vars0d), mesh.cache, mesh,
                        joinpath(outputdir, "output0d"),
                        savecb_timing;
                        format0d=prms["Output"]["format0d"])
  vars1d = prms["Output"]["variables1d"]
  savecb_1d = SaveCallback(Symbol.(vars1d), mesh.cache, mesh,
                        joinpath(outputdir, "output1d"),
                        savecb_timing;
                        format1d=prms["Output"]["format1d"])
  vars2d = prms["Output"]["variables2d"]
  savecb_2d = SaveCallback(Symbol.(vars2d), mesh.cache, mesh,
                        joinpath(outputdir, "output2d"),
                        savecb_timing;
                        format2d=prms["Output"]["format2d"])
  savecbs = [ cb for cb in (savecb_0d, savecb_1d, savecb_2d) if !isnothing(cb) ]

  intrpcb_timing = CallbackTiming(every_dt=prms["Output"]["interpolate_every_dt"],
                                  every_iteration=prms["Output"]["interpolate_every_iteration"])
  intrpcb = InterpolationCallback(Symbol.(prms["Output"]["interpolate_variables"]), mesh,
                                  prms["Output"]["interpolate_nodes"],
                                  joinpath(outputdir, "interpolate"),
                                  intrpcb_timing;
                                  format1d="h5")

  wrapcb = FunctionCallback(wrap_dynamic, CallbackTiming(every_iteration=1))
  filtercb = if prms["Mesh"]["filter_enable"]
    FunctionCallback(cb_spectral_filter!, CallbackTiming(every_iteration=1))
  else
    nothing
  end

  # TODO This leads to incorrect outputs in case interpolate_aligned_ts overwrites
  # the proposed timestep of timealigned_savecb
  aligned_ts = prms["Output"]["aligned_ts"]
  timealigned_savecb, wrapped_timestep! = if length(aligned_ts) > 0 && length(savecbs) > 0
    atscb = TimeAlignedSaveCallback(aligned_ts, wrapped_timestep!, savecbs)
    atscb, atscb.aligned_timestep_fn
  else
    nothing, wrapped_timestep!
  end
  interpolate_aligned_ts = prms["Output"]["interpolate_aligned_ts"]
  timealigned_intrpcb, wrapped_timestep! = if length(interpolate_aligned_ts) > 0 && !isnothing(intrpcb)
    atscb = TimeAlignedSaveCallback(interpolate_aligned_ts, wrapped_timestep!, intrpcb)
    atscb, atscb.aligned_timestep_fn
  else
    nothing, wrapped_timestep!
  end

  timercb = TimerOutputCallback(TO, joinpath(outputdir, "timer.log"),
              CallbackTiming(every_dt_walltime=prms["Output"]["timer_every_dt_walltime"]))

  pcb = ProgressCallback(prms["Evolution"]["tend"],
                         prms["Output"]["progress_every_dt_walltime"], TO,
                         verbose=prms["Log"]["progress_stdout"])

  # TODO Shouldn't filtercb and wrapcb be swapped?
  pushfirst!(env.callbacks, filtercb)
  pushfirst!(env.callbacks, wrapcb)
  append!(env.callbacks, savecbs)
  push!(env.callbacks, timealigned_savecb)
  push!(env.callbacks, intrpcb)
  push!(env.callbacks, timealigned_intrpcb)
  push!(env.callbacks, timercb)
  push!(env.callbacks, pcb)

  pushfirst!(env.callbacks_substep, wrapcb)

  substep_savecb_timing = CallbackTiming(every_dt=prms["Output"]["substep_every_dt"],
                                         every_iteration=prms["Output"]["substep_every_iteration"]);
  substep_savecb = SaveCallback(Symbol.(prms["Output"]["substep_variables"]), mesh.cache, mesh,
                                joinpath(outputdir, "substep_output"),
                                substep_savecb_timing,
                                format1d="h5")
  if !isnothing(substep_savecb)
    remote_substep_savecb = convert_to_remote_savecallback!(substep_savecb)
    append!(env.callbacks, remote_substep_savecb)
    push!(env.callbacks_substep, substep_savecb)
  end


  return wrapped_rhs!, wrapped_timestep!
end


regrid!(mesh, cache, evolution) = nothing


function evolve!(env, rhs_fn, timestep_fn, prms)

  @unpack tend, method = prms["Evolution"]
  @unpack mesh, cache, callbacks, callbacks_substep = env

  # TODO This should be hidden behind an interface
  u0 = reduce(vcat, fields(cache.dynamic_variables))
  alg = algorithm(Symbol(method))
  evolution = Evolution(rhs_fn, u0, timestep_fn, tend, alg; mesh=mesh,
                        callback_fullstep=callbacks, callback_substep=callbacks_substep)

  # run once on initial data
  callbacks(u0, evolution.t, evolution.it, force=true)

  t_last_amr, amr_every_dt = Inf, Inf
  # t_last_amr = amr_every_dt
  while !done(evolution)
    success = step!(evolution)
    if success != true
      break
    end
    if evolution.t >= t_last_amr
      regrid!(mesh, cache, evolution)
      t_last_amr += round(( (evolution.t - t_last_amr) / amr_every_dt )) * amr_every_dt
    end
  end

end
